/*
 *  OverviewManagerImpl.scala
 *  (Overview)
 *
 *  Copyright (c) 2010-2013 Hanns Holger Rutz. All rights reserved.
 *
 *	This software is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either
 *	version 2, june 1991 of the License, or (at your option) any later version.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public
 *	License (gpl.txt) along with this software; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *	For further information, please contact Hanns Holger Rutz at
 *	contact@sciss.de
 *    28-Mar-10   extracted from Kontur. Removing app package dependancy
 */

package de.sciss.sonogram
package impl

import java.awt.image.BufferedImage
import java.io.File
import de.sciss.dsp.ConstQ
import de.sciss.synth.io.{AudioFile, Frames}
import de.sciss.sonogram.ResourceManager.{Image, ImageSpec}
import de.sciss.model.impl.ModelImpl
import de.sciss.filecache.{Limit, Producer}

private object OverviewManagerImpl {
  private final class ConstQCache(val constQ: ConstQ) {
    var useCount = 0
  }

  private final class FileBufCache(val buf: Frames) {
    var useCount = 0
  }

  private final class ImageCache(val img: BufferedImage) {
    var useCount = 0
  }

  private final class OverviewCache(val overview: OverviewImpl) {
    var useCount = 0
  }
}
private[sonogram] final class OverviewManagerImpl(val config: OverviewManager.Config)
  extends OverviewManager with ModelImpl[OverviewManager.Update] with ResourceManager {

  import OverviewManager._
  import OverviewManagerImpl._
  import Overview.{Config => OvrSpec, Output => OvrOut}
  import OverviewImpl.debug

  //  /** Creates the file for the overview cache meta data. It should have a filename extension,
  //    * and the manager will store the overview binary data in a file with the same prefix but
  //    * different filename extension.
  //    *
  //    * @param path   the audio input file to derive the cache from
  //    */
  //  protected def createCacheFile(path: File): File

  /** This is a constant, but can be overridden by subclasses. */
  protected val decimation  = List(1, 6, 6, 6, 6)

  private var constQCache   = Map.empty[SonogramSpec, ConstQCache]
  private var imageCache    = Map.empty[(Int, Int), ImageCache]
  private var fileBufCache  = Map.empty[ImageSpec, FileBufCache]

  // private var futures       = Map.empty[OvrSpec, Future[OvrOut]]
  private var overviewCache = Map.empty[OvrSpec, OverviewCache]
  private val sync          = new AnyRef

  private val producer = {
    val cc = Producer.Config[OvrSpec, OvrOut]()
    cc.extension  = "sono"
    cc.accept     = { (spec: OvrSpec, out: OvrOut) =>
      val specAF = AudioFile.readSpec(spec.file)
      debug(s"accept ${spec.file.lastModified()} == ${out.input.lastModified} && $specAF == ${out.input.fileSpec}")
      spec.file.lastModified() == out.input.lastModified && specAF == out.input.fileSpec
    }
    cc.space      = { (spec: OvrSpec, out: OvrOut) =>
      val res = out.output.length()
      debug(s"space of $out is $res")
      res
    }
    cc.evict      = { (spec: OvrSpec, out: OvrOut) =>
      debug(s"evict $out")
      out.output.delete()
    }
    cc.executionContext = config.executionContext

    config.caching match {
      case Some(c) =>
        cc.capacity = Limit(space = c.sizeLimit)
        debug(s"Producer capacity is ${cc.capacity}")
        cc.folder   = c.folder

      case _ =>
        cc.capacity = Limit(count = 0, space = 0L)
        val f = File.createTempFile(".cache", "")
        f.delete()
        f.mkdir()
        f.deleteOnExit()
        cc.folder   = f
    }
    Producer(cc)
  }

  //  private val consumer = Consumer(producer) { ovrSpec =>
  //    overviews
  //    ... : Future[OvrOut]
  //  }

  def acquire(job: Job): Overview = {
    val (ovrSpec, ovrIn) = jobToOverviewConfig(job)
    sync.synchronized {
      val entry = overviewCache.getOrElse(ovrSpec, {
        val res = new OverviewCache(new OverviewImpl(ovrSpec, ovrIn, manager = this, producer = producer))
        overviewCache += ovrSpec -> res
        res
      })
      entry.useCount += 1
      entry.overview
    }
  }

  def release(overview: Overview) {
    val ovrSpec = overview.config
    sync.synchronized {
      val entry = overviewCache.getOrElse(ovrSpec,
        throw new IllegalStateException(s"Trying to release an unregistered overview")
      )
      entry.useCount -= 1
      debug(s"release $overview; count = ${entry.useCount}")
      if (entry.useCount == 0) {
        overviewCache -= ovrSpec
        entry.overview.dispose()
      }
    }
  }

  private def jobToOverviewConfig(job: Job): (Overview.Config, Overview.Input) = {
    val inSpec      = AudioFile.readSpec(job.file)
    val sampleRate  = inSpec.sampleRate
    val cq          = job.analysis
    val stepSize    = (cq.maxTimeRes/1000 * sampleRate + 0.5).toInt

    val sonogram    = SonogramSpec(
      sampleRate = sampleRate, minFreq = cq.minFreq,
      maxFreq = math.min(cq.maxFreq, sampleRate / 2).toFloat, bandsPerOct = cq.bandsPerOct,
      maxFFTSize = cq.maxFFTSize, stepSize = stepSize
    )

    (Overview.Config(job.file, sonogram, decimation), Overview.Input(inSpec, job.file.lastModified()))
  }

  def dispose() {
    sync.synchronized {
      releaseListeners()
      overviewCache.foreach(_._2.overview.abort())
    }
  }

  def allocateConstQ(spec: SonogramSpec): ConstQ = {
    sync.synchronized {
      val entry = constQCache.get(spec) getOrElse {
        val res = new ConstQCache(constQFromSpec(spec))
        constQCache += spec -> res
        res
      }
      entry.useCount += 1
      entry.constQ
    }
  }

  def releaseConstQ(spec: SonogramSpec) {
    sync.synchronized {
      val entry = constQCache(spec) // let it throw an exception if not contained
      entry.useCount -= 1
      if (entry.useCount == 0) {
        constQCache -= spec
      }
    }
  }

  def allocateImage(spec: ImageSpec): Image = {
    sync.synchronized {
      val img     = allocateImage(spec.width, spec.height)
      val fileBuf = allocateFileBuf(spec)
      new Image(img, fileBuf)
    }
  }

  def releaseImage(spec: ImageSpec) {
    sync.synchronized {
      releaseImage(spec.width, spec.height)
      releaseFileBuf(spec)
    }
  }

  private def allocateImage(width: Int, height: Int): BufferedImage = {
    sync.synchronized {
      val entry = imageCache.getOrElse((width, height), {
        val res = new ImageCache(new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB))
        imageCache += (width, height) -> res
        res
      })
      entry.useCount += 1
      entry.img
    }
  }

  private def allocateFileBuf(spec: ImageSpec): Array[Array[Float]] = {
    sync.synchronized {
      val entry = fileBufCache.getOrElse(spec, {
        val res = new FileBufCache(Array.ofDim[Float](spec.numChannels, spec.width * spec.height))
        fileBufCache += spec -> res
        res
      })
      entry.useCount += 1
      entry.buf
    }
  }

  private def releaseImage(width: Int, height: Int) {
    sync.synchronized {
      val key         = (width, height)
      val entry       = imageCache(key) // let it throw an exception if not contained
      entry.useCount -= 1
      if (entry.useCount == 0) {
        imageCache -= key
      }
    }
  }

  private def releaseFileBuf(spec: ImageSpec) {
    sync.synchronized {
      val entry = fileBufCache(spec) // let it throw an exception if not contained
      entry.useCount -= 1
      if (entry.useCount == 0) {
        fileBufCache -= spec
      }
    }
  }

  private def constQFromSpec(spec: SonogramSpec): ConstQ = {
    val cfg = ConstQ.Config()
    cfg.sampleRate  = spec.sampleRate
    cfg.minFreq     = spec.minFreq
    cfg.maxFreq     = spec.maxFreq
    cfg.bandsPerOct = spec.bandsPerOct
    val maxTimeRes  = spec.stepSize / spec.sampleRate * 1000
    //    cfg.maxTimeRes  = spec.maxTimeRes
    cfg.maxTimeRes  = maxTimeRes.toFloat  // note: this is a purely informative field
    cfg.maxFFTSize  = spec.maxFFTSize
    ConstQ(cfg)
  }
}